import 'dart:async';
import 'package:flutter/foundation.dart';
import 'package:huddle01_flutter/src/event_emitter.dart';
import 'package:flutter_webrtc/flutter_webrtc.dart';
import 'package:huddle01_flutter/src/peer.dart';
import 'package:huddle01_flutter/src/web_socket.dart';
import 'package:mediasoup_client_flutter/mediasoup_client_flutter.dart';

class HuddleClientRepository {
  List<Consumer> _consumers = [];
  Map<String, Producer> _producers = Map();
  List<Peer> _peers = [];

  final String roomId;
  final String peerId;
  final String url;
  final String displayName;

  bool _closed = false;

  WebSocket? _webSocket;
  Device? _mediasoupDevice;
  Transport? _sendTransport;
  Transport? _recvTransport;
  bool _produce = false;
  bool _consume = true;
  String audioInputDeviceId;
  late String audioOutputDeviceId;
  String videoInputDeviceId;

  HuddleClientRepository({
    required this.roomId,
    required this.peerId,
    required this.url,
    required this.displayName,
    required this.audioInputDeviceId,
    required this.videoInputDeviceId,
  });

  void close() {
    if (_closed) {
      return;
    }

    _webSocket?.close();
    _sendTransport?.close();
    _recvTransport?.close();
  }

  Future<void> disableMic() async {
    String micId = _producers['mic']!.id;
    _producers['mic']?.close();
    emitter.emit('removeProducer', 'mic');
    try {
      await _webSocket!.socket.request('closeProducer', {
        'producerId': micId,
      });
    } catch (error) {}
  }

  Future<void> disableWebcam() async {
    String webcamId = _producers['webcam']!.id;
    _producers['webcam']?.close();
    emitter.emit('removeProducer', 'webcam');

    try {
      await _webSocket!.socket.request('closeProducer', {
        'producerId': webcamId,
      });
    } catch (error) {
    } finally {}
  }

  Future<void> muteMic() async {
    _producers['mic']?.pause();

    try {
      await _webSocket!.socket.request('pauseProducer', {
        'producerId': _producers['mic']!.id,
      });
    } catch (error) {}
  }

  Future<void> unmuteMic() async {
    _producers['mic']!.resume();

    try {
      await _webSocket!.socket.request('resumeProducer', {
        'producerId': _producers['mic']!.id,
      });
    } catch (error) {}
  }

  void _producerCallback(Producer producer) {
    if (producer.source == 'mic') {
      producer.on('transportclose', () {
        emitter.emit('removeProducer', 'mic');
        _producers['mic']?.close();
      });

      producer.on('trackended', () {
        disableMic().catchError((data) {});
      });
    } else if (producer.source == 'webcam') {
      producer.on('transportclose', () {
        emitter.emit('removeProducer', 'webcam');
        _producers['webcam']?.close();
      });

      producer.on('trackended', () {
        disableWebcam().catchError((data) {});
      });
      // meBloc.add(MeSetWebcamInProgress(progress: false));
    }

    emitter.emit('addProducer', producer);
    _producers[producer.source] = producer;
  }

  void _consumerCallback(Consumer consumer, [dynamic accept]) {
    consumer.on('transportclose', () {
      // consumersBloc.add(ConsumerRemove(consumerId: consumer.id));
      // emitter.emit('removeConsumer', consumer.id);
      // _consumers.remove(consumer);
    });

    ScalabilityMode scalabilityMode = ScalabilityMode.parse(
        consumer.rtpParameters.encodings.first.scalabilityMode);

    accept({});

    // consumersBloc.add(ConsumerAdd(consumer: consumer));
    emitter.emit('addConsumer', consumer);
    _consumers.add(consumer);
    // peersBloc
    //     .add(PeerAddConsumer(peerId: consumer.peerId, consumerId: consumer.id));
    emitter.emit('addPeerConsumer', consumer);
    _peers
        .firstWhere((element) => element.id == consumer.peerId)
        .consumers
        .add(consumer.id);
  }

  Future<MediaStream> createAudioStream() async {
    // audioInputDeviceId = mediaDevicesBloc.state.selectedAudioInput.deviceId;
    Map<String, dynamic> mediaConstraints = {
      'audio': {
        'optional': [
          {
            'sourceId': audioInputDeviceId,
          },
        ],
      },
    };

    MediaStream stream =
        await navigator.mediaDevices.getUserMedia(mediaConstraints);

    return stream;
  }

  Future<MediaStream> createVideoStream() async {
    // videoInputDeviceId = mediaDevicesBloc.state.selectedVideoInput.deviceId;
    Map<String, dynamic> mediaConstraints = <String, dynamic>{
      'audio': false,
      'video': {
        'mandatory': {
          'minWidth':
              '1280', // Provide your own width, height and frame rate here
          'minHeight': '720',
          'minFrameRate': '30',
        },
        'optional': [
          {
            'sourceId': videoInputDeviceId,
          },
        ],
      },
    };

    MediaStream stream =
        await navigator.mediaDevices.getUserMedia(mediaConstraints);

    return stream;
  }

  void enableWebcam() async {
    // meBloc.add(MeSetWebcamInProgress(progress: true));
    if (_mediasoupDevice!.canProduce(RTCRtpMediaType.RTCRtpMediaTypeVideo) ==
        false) {
      return;
    }
    MediaStream? videoStream;
    MediaStreamTrack track;
    try {
      RtpCodecCapability? codec = _mediasoupDevice!.rtpCapabilities.codecs
          .firstWhere(
              (RtpCodecCapability c) => c.mimeType.toLowerCase() == 'video/vp9',
              orElse: () =>
                  throw 'desired vp9 codec+configuration is not supported');
      videoStream = await createVideoStream();
      track = videoStream.getVideoTracks().first;
      // meBloc.add(MeSetWebcamInProgress(progress: true));
      _sendTransport!.produce(
        track: track,
        codecOptions: ProducerCodecOptions(
          videoGoogleStartBitrate: 1000,
        ),
        encodings: kIsWeb
            ? [
                RtpEncodingParameters(scalabilityMode: 'S3T3_KEY'),
              ]
            : [],
        stream: videoStream,
        appData: {
          'source': 'webcam',
        },
        source: 'webcam',
        codec: codec,
      );
    } catch (error) {
      if (videoStream != null) {
        await videoStream.dispose();
      }
    }
  }

  void enableMic() async {
    if (_mediasoupDevice!.canProduce(RTCRtpMediaType.RTCRtpMediaTypeAudio) ==
        false) {
      return;
    }

    MediaStream? audioStream;
    MediaStreamTrack track;
    try {
      audioStream = await createAudioStream();
      track = audioStream.getAudioTracks().first;
      _sendTransport!.produce(
        track: track,
        codecOptions: ProducerCodecOptions(opusStereo: 1, opusDtx: 1),
        stream: audioStream,
        appData: {
          'source': 'mic',
        },
        source: 'mic',
      );
    } catch (error) {
      if (audioStream != null) {
        await audioStream.dispose();
      }
    }
  }

  Future<void> _joinRoom() async {
    try {
      _mediasoupDevice = Device();

      dynamic routerRtpCapabilities =
          await _webSocket!.socket.request('getRouterRtpCapabilities', {});

      print(routerRtpCapabilities);

      final rtpCapabilities = RtpCapabilities.fromMap(routerRtpCapabilities);
      rtpCapabilities.headerExtensions
          .removeWhere((he) => he.uri == 'urn:3gpp:video-orientation');
      await _mediasoupDevice!.load(routerRtpCapabilities: rtpCapabilities);

      if (_mediasoupDevice!.canProduce(RTCRtpMediaType.RTCRtpMediaTypeAudio) == true ||
          _mediasoupDevice!.canProduce(RTCRtpMediaType.RTCRtpMediaTypeVideo) == true) {
        _produce = true;
      }

      if (_produce) {
        Map transportInfo =
            await _webSocket!.socket.request('createWebRtcTransport', {
          'forceTcp': false,
          'producing': true,
          'consuming': false,
          'sctpCapabilities': _mediasoupDevice!.sctpCapabilities.toMap(),
        });

        _sendTransport = _mediasoupDevice!.createSendTransportFromMap(
          transportInfo,
          producerCallback: _producerCallback,
        );

        _sendTransport!.on('connect', (Map data) {
          _webSocket!.socket
              .request('connectWebRtcTransport', {
                'transportId': _sendTransport!.id,
                'dtlsParameters': data['dtlsParameters'].toMap(),
              })
              .then(data['callback'])
              .catchError(data['errback']);
        });

        _sendTransport!.on('produce', (Map data) async {
          try {
            Map response = await _webSocket!.socket.request(
              'produce',
              {
                'transportId': _sendTransport!.id,
                'kind': data['kind'],
                'rtpParameters': data['rtpParameters'].toMap(),
                if (data['appData'] != null)
                  'appData': Map<String, dynamic>.from(data['appData'])
              },
            );

            data['callback'](response['id']);
          } catch (error) {
            data['errback'](error);
          }
        });

        _sendTransport!.on('producedata', (data) async {
          try {
            Map response = await _webSocket!.socket.request('produceData', {
              'transportId': _sendTransport!.id,
              'sctpStreamParameters': data['sctpStreamParameters'].toMap(),
              'label': data['label'],
              'protocol': data['protocol'],
              'appData': data['appData'],
            });

            data['callback'](response['id']);
          } catch (error) {
            data['errback'](error);
          }
        });
      }

      if (_consume) {
        Map transportInfo = await _webSocket!.socket.request(
          'createWebRtcTransport',
          {
            'forceTcp': false,
            'producing': false,
            'consuming': true,
            'sctpCapabilities': _mediasoupDevice!.sctpCapabilities.toMap(),
          },
        );

        _recvTransport = _mediasoupDevice!.createRecvTransportFromMap(
          transportInfo,
          consumerCallback: _consumerCallback,
        );

        _recvTransport!.on(
          'connect',
          (data) {
            _webSocket!.socket
                .request(
                  'connectWebRtcTransport',
                  {
                    'transportId': _recvTransport!.id,
                    'dtlsParameters': data['dtlsParameters'].toMap(),
                  },
                )
                .then(data['callback'])
                .catchError(data['errback']);
          },
        );
      }

      Map response = await _webSocket!.socket.request('join', {
        'displayName': displayName,
        'device': {
          'name': "Flutter",
          'flag': 'flutter',
          'version': '0.4.3',
        },
        'rtpCapabilities': _mediasoupDevice!.rtpCapabilities.toMap(),
        'sctpCapabilities': _mediasoupDevice!.sctpCapabilities.toMap(),
      });

      response['peers'].forEach((value) {
        // peersBloc.add(PeerAdd(newPeer: value));
        _peers.add(Peer.fromMap(value));
        emitter.emit('addPeer', value);
      });

      if (_produce) {
        enableMic();
        enableWebcam();

        _sendTransport!.on('connectionstatechange', (connectionState) {
          if (connectionState == 'connected') {
            // enableChatDataProducer();
            // enableBotDataProducer();
          }
        });
      }
    } catch (error) {
      print(error);
      close();
    }
  }

  void join() {
    _webSocket = WebSocket(
      peerId: peerId,
      roomId: roomId,
      url: url,
    );

    _webSocket!.onOpen = _joinRoom;
    _webSocket!.onFail = () {
      print('WebSocket connection failed');
    };
    _webSocket!.onDisconnected = () {
      if (_sendTransport != null) {
        _sendTransport!.close();
        _sendTransport = null;
      }
      if (_recvTransport != null) {
        _recvTransport!.close();
        _recvTransport = null;
      }
    };
    _webSocket!.onClose = () {
      if (_closed) return;

      close();
    };

    _webSocket!.onRequest = (request, accept, reject) async {
      switch (request['method']) {
        case 'newConsumer':
          {
            if (!_consume) {
              reject(403, 'I do not want to consume');
              break;
            }
            try {
              _recvTransport!.consume(
                id: request['data']['id'],
                producerId: request['data']['producerId'],
                kind: RTCRtpMediaTypeExtension.fromString(
                    request['data']['kind']),
                rtpParameters:
                    RtpParameters.fromMap(request['data']['rtpParameters']),
                appData: Map<String, dynamic>.from(request['data']['appData']),
                peerId: request['data']['peerId'],
                accept: accept,
              );
            } catch (error) {
              print('newConsumer request failed: $error');
              throw (error);
            }
            break;
          }
        default:
          break;
      }
    };

    _webSocket!.onNotification = (notification) async {
      switch (notification['method']) {
        //TODO: todo;
        case 'producerScore':
          {
            break;
          }
        case 'consumerClosed':
          {
            String consumerId = notification['data']['consumerId'];
            // consumersBloc.add(ConsumerRemove(consumerId: consumerId));
            emitter.emit('removeConsumer', consumerId);
            _consumers
                .firstWhere((element) => element.id == consumerId)
                .close();
            _consumers.removeWhere((element) => element.id == consumerId);
            // peersBloc.add(PeerRemoveConsumer(
            //     peerId: notification['data']['peerId'],
            //     consumerId: consumerId));
            emitter.emit('peerRemoveConsumer', notification['data']['peerId'],
                consumerId);
            _peers[notification['data']['peerId']]
                .consumers
                .removeWhere((element) => element == consumerId);

            break;
          }
        case 'consumerPaused':
          {
            String consumerId = notification['data']['consumerId'];
            // consumersBloc.add(ConsumerPaused(consumerId: consumerId));
            _consumers
                .firstWhere((element) => element.id == consumerId)
                .pause();
            break;
          }

        case 'consumerResumed':
          {
            String consumerId = notification['data']['consumerId'];
            // consumersBloc.add(ConsumerResumed(consumerId: consumerId));
            _consumers
                .firstWhere((element) => element.id == consumerId)
                .resume();
            break;
          }

        case 'newPeer':
          {
            Map newPeer = Map<String, dynamic>.from(notification['data']);
            // peersBloc.add(PeerAdd(newPeer: newPeer));
            _peers.add(Peer.fromMap(newPeer));
            emitter.emit('addPeer', newPeer);
            break;
          }

        case 'peerClosed':
          {
            String peerId = notification['data']['peerId'];
            // peersBloc.add(PeerRemove(peerId: peerId));
            _peers.removeWhere((element) => element.id == peerId);
            emitter.emit('removePeer', peerId);
            break;
          }

        default:
          break;
      }
    };
  }
}
